//
//  maxlib.c
//
//  Created by Orly Natan on 22/11/17.
//

#include "maxlib.h"


bool listenOn = false;

// continuously listen to board and update info_pins values
void maxclient_listen(board_t *board) {
    listenOn = true;
    while(listenOn) {
        firmataclient_onidle(board);
    }
}

// connect to board via a serial port, init board to requested configuration
// and start listening to updates from the board
board_t* maxclient_newboard(char *name, int baud, char *err) {
    
    // allocate memory for new board
    board_t *board = (board_t*)malloc(sizeof(board_t));
    if(board == NULL) {
        strcpy(err, "firmataclient_initport: couldn't allocate new memory");
        return NULL;
    }
    
    // open connection to the board, and request REPORT_FIRMATA
    bool result = firmataclient_initport(board, name, baud);
    if (result == false) {
        sprintf(err, "firmataclient_initport: couldn't open port: %s, %s", board->err_msg, board->port->error_msg);
        return board;
    }
    
    return board;
}

// connect to board via a serial port, init board to requested configuration
// and start listening to updates from the board
int maxclient_connect(board_t *board, int len_config, int *pins, int *modes, int func) {
    
    if (board == NULL) return -1;
    
    // request firmware report
    firmataclient_requestFirmware(board);
    
    // wait for board to be ready for communication
    firmataclient_waitForBoardReady(board);
    
    // if it's a READ (boardin) connection, then
    // launch infine listener thread
    if(func == IN_BOARD) {
        // start new thread for (continuous) listening to port updates
        // should be *before* requesting FIRMATA_REPORT
        bool result = firmataclient_launchtlisteninghread(board);
        if(result != 0) {
            sprintf(board->err_msg, "firmataclient_launchtlisteninghread: %s", board->err_msg);
            return result;
        }
    }
    
    // now that the board is ready for communication
    // set the correct pin configuration as requested (if supported),
    // or default is MODE_OFF (255)
    maxclient_configurePins(board, len_config, pins, modes);
    
    // now the max object is ready for work.
    return 0;
}

int maxclient_disconnect(board_t *board) {
    
    if (board == NULL) return -1;
    
    // for a READ (boardin) connection, kill listener thread
    if(firmataclient_destroylisteningthread(board) == -1) {
        sprintf(board->err_msg, "maxclient_disconnect: %s", board->err_msg);
        return -1;
    }
    else
        return 0;
}

// update board with requested configuration
void maxclient_configurePins(board_t *board, int len_config, int *pins, int *modes) {
    
    // note: assumes that the board is ready for communication.
    
    // set the correct pin configuration as requested (if supported),
    // or default is MODE_OFF (255)
    if(len_config > 128) len_config = 128;
    for(int i=0; i<len_config; ++i) {
        firmataclient_changemode(board, pins[i], modes[i]);
    }
}

// pause the listener from receiving incoming board messages
void maxclient_pause(void) {
    listenOn = false;
}

// restart listening to board after pausing (unchanged configuration) ???
void maxclient_restart(void) {
    listenOn = true;
}

// stop communication with the board via the serial port and disconnect
void maxclient_stop(board_t *board) {

    listenOn = false;

    // if no further objects are connected to board,
    // then close the port
    firmataclient_closeport(board);
}


void maxclient_moveServo(board_t *board, int pin, int newVal) {
    // check pin index
    if (pin < 0 || pin > TOTAL_IN_PINS) return;
    // limit newVal to servo range
    if (newVal < 0) newVal = 0;
    if (newVal > 170) newVal = 170;
    // scale, if needed
    //newVal = (newVal * 180) / 1023;
    // update the board
    firmataclient_setnewvalue(board, pin, newVal);
}

void maxclient_movePwm(board_t *board, int pin, int newVal) {
    // check pin index
    if (pin < 0 || pin > TOTAL_IN_PINS) return;
    // limit newVal range
    if (newVal < 0) newVal = 0;
    if (newVal > 170) newVal = 170;
    // update the board
    firmataclient_setnewvalue(board, pin, newVal);
}

void maxclient_toggleDigital(board_t *board, int pin, int newVal) {
    // check pin index
    if (pin < 0 || pin > TOTAL_IN_PINS) return;
    // limit newVal to binary 0 or 1
    newVal = (newVal <= 0? 0: 1);
    // update the board
    firmataclient_setdigitalvalue(board, pin, newVal);
}

void maxclient_changeMode(board_t *board, int pin, int mode) {
    
    // todo: check the mode is supported..
    
    // update the board
    firmataclient_changemode(board, pin, mode);
}

void maxclient_toggleListener(board_t *board) {
    // listenOn = !listenOn;
    if(listenOn) {
        //pause
        listenOn = false;
    }
    else {
        listenOn = true;
        maxclient_listen(board);
    }
}

void maxclient_listenerOn(void) {
        listenOn = true;
}

void maxclient_listenerOff(void) {
        listenOn = false;
}

void maxclient_getInputValues(board_t *board, int *indexArray,int *valuesArray, int pinsCount) {
    firmataclient_getInputValues(board, indexArray, valuesArray, pinsCount);
}

int maxclient_shutdown(board_t *board) {
    return firmataclient_closeport(board);
}



